
#include "GoalStrategy.h"
#include "Game.h"
#include <algorithm>
#include "Random.h"

#include "bighashmap.cc"
template class BigHashMap<Coord, unsigned>;

int
hashcode (const Coord& p) {
  unsigned p1 = hashcode(p.first);
  unsigned p2 = hashcode(p.second);
  return p1 & ((p2 << 8) | (p2 >> 24));
}

GoalStrategy::GoalStrategy(Game& g) : _game(g), _depth(300) {}

void
GoalStrategy::set_depth(int d) { _depth = d; }

void 
GoalStrategy::get_goal()
{
  calculate_distances();

  _game._ttime.load_distances(_game._me._position);
  choose_packages();
  unsigned u;
  choose_goal(_game._me, _game._cached_goal, u);

  HomeBase* hb;
  if ((hb = _game._board._homebases.findp(_game._me._position)) != NULL) {
    hb->_last_value = (int)u - (int)heuristic_base(_game._me);
    hb->_visited = true;
  }
}

void
GoalStrategy::calculate_distances()
{
  _distancemap.clear();
  _points.clear();
  _packages.clear();

  // init the packages vector
  list<unsigned>::iterator i;
  for (i = _game._me._carrying.begin(); i != _game._me._carrying.end(); i++)
    _packages.push_back(*i);
  for (i = _game._standingOn.begin(); i != _game._standingOn.end(); i++)
    _packages.push_back(*i);

  // add my own point
  _points.push_back(_game._me._position);

  // add dest points
  vector<unsigned>::iterator j;
  for (j = _packages.begin(); j != _packages.end(); j++) 
    _points.push_back(_game._packages[*j]._dest);

  // add base points
  BigHashMap<Point, HomeBase>::Iterator k(&_game._board._homebases);
  for (; k; k++) 
    _points.push_back(k.key());

  // TODO - make sure there are no redundant entries

  vector<Point>::iterator m, n;
  for (m = _points.begin(); m != _points.end(); m++) {
    _game._ttime.load_distances(*m);

    for(n = _points.begin(); n != _points.end(); n++) {
      _distancemap.find_force(Coord(*m,*n)).push_back(
	DistanceVal(Coord(*m,*n),_game._ttime.get_distance_to(*n)));
    }
  }
}

void
GoalStrategy::choose_packages()
{
  _game._packages_goal = _game._me._carrying;
  if (_game._standingOn.size() ==0) {
    return;
  }

  unsigned maxscore = 0;
  if (_game._me._carrying.size() > 0)
    maxscore = heuristic(_game._me) +
      200*_game._stats.points_per_hundred_dollars(); 
  // it takes two turns to pick/drop

  // generate random sets
  Player temp = _game._me;
  list<unsigned> best = _game._me._carrying;

  for (int j = 0; j < 20; j++) {
    get_random_subset(temp._carrying, Player::_bagsize);
    unsigned score = heuristic(temp);
    if (score > maxscore) {
      maxscore = score;
      best = temp._carrying;
    }
  }
  _game._packages_goal = best;
}

void
GoalStrategy::choose_goal(const Player &p, Point &goal, unsigned &maxscore) {
  maxscore = 0;
  unsigned score;
  Player temp = p;
  goal = p._position;

  // check the scores of package destinations
  list<unsigned>::iterator i;
  for (i = _game._packages_goal.begin(); 
      i != _game._packages_goal.end(); i++ )
  {
    if (_game._packages[*i]._dest == p._position)
      continue; // not useful to go where you are
    temp._position = _game._packages[*i]._dest;
    temp._dollars -= get_distance(p._position, temp._position)/100;
    score = heuristic(temp);
    if (score > maxscore)
    {
      maxscore = score;
      goal = temp._position;
    }
    temp._dollars += get_distance(p._position, temp._position)/100;
  }
  
  unsigned weight = 0;
  for (i = temp._carrying.begin(); i != temp._carrying.end(); i++) {
    weight += _game._packages[*i]._weight;
  }

  // check the scores of going to the home-bases
  BigHashMap<Point, HomeBase>::Iterator k(&_game._board._homebases);
  for (; k; k++)  {
    if (k.key() == p._position)
      continue;// not useful to go where you are

    temp._position = k.key();
    temp._dollars -= get_distance(p._position, temp._position)/100;
    int extra_points;
    if (k.value()._visited) 
      extra_points = k.value()._last_value;
    else
      extra_points = _game.new_homebase_points();

    score = heuristic_base(temp) + extra_points;
    if (score > maxscore)
    {
      maxscore = score;
      goal = temp._position;
    }
    temp._dollars += get_distance(p._position, temp._position)/100;
  }
}

unsigned
GoalStrategy::heuristic_base(const Player &player)
{
  return player._points * 100 
    + player._dollars * _game._stats.points_per_hundred_dollars() ;
}
unsigned
GoalStrategy::heuristic(const Player &player)
{
  unsigned base = heuristic_base(player);

  for (list<unsigned>::const_iterator i = player._carrying.begin();
      i != player._carrying.end(); i++) 
  {
    int value = _game._packages[*i]._weight*100 - 
      get_distance(player._position, _game._packages[*i]._dest) * 
      _game._stats.points_per_hundred_dollars()/100;
    if (value > 3)
      base += value;
    else 
      base += 3;
  }
  return base;
}
unsigned
GoalStrategy::get_distance(Point p1, Point p2){
    Coord c(p1, p2);
    vector<DistanceVal> *dv =_distancemap.findp(c);
    return find(dv->begin(), dv->end(), DistanceVal(c,0))->_dist;
}

void
GoalStrategy::get_random_subset(list<unsigned> &packages, unsigned maxweight) {
  vector<unsigned> allpacks;
  vector<bool> in;
  unsigned weight = 0;
  list<unsigned>::iterator i;
  for (i = _game._me._carrying.begin();
      i != _game._me._carrying.end(); i++) {
      allpacks.push_back(*i);
      in.push_back(false);
  }
  for (i = _game._standingOn.begin();
      i != _game._standingOn.end(); i++) {
      allpacks.push_back(*i);
      in.push_back(false);
  }

  int newpack;
  Point center_mass;
  int numItems;
  if ((newpack = choose_random_package(allpacks, in, weight, maxweight)) != -1)
  {
    complete_set(_game._packages[newpack]._dest, allpacks, in, weight, 
	maxweight, center_mass, numItems);
    while ((newpack = choose_nearby_package(allpacks, in, weight, maxweight,
	    center_mass, numItems)) != -1)
    {
      complete_set(_game._packages[newpack]._dest, allpacks, in, weight, 
	  maxweight, center_mass, numItems);
    }
  }

  /*
  while (weight > maxweight) {
    j = rand_int(allpacks.size());
    while (in[j] == false)
      j = (j + 1) % allpacks.size();
    in[j] = false;
      weight -= _game._packages[allpacks[j]]._weight;
  }
  */

  unsigned j;
  packages.clear();
  for (j = 0; j < allpacks.size(); j++) {
    if (in[j])
      packages.push_back(allpacks[j]);
  }
}

int
GoalStrategy::choose_nearby_package( const vector<unsigned> &allpacks,
    const vector<bool> &in, unsigned weight, unsigned maxweight, 
    Point sum, int numItems)
{
  int mindist = 10000000;
  int minindex = -1;
  int dist;
  unsigned i;
  for (i = 0; i < allpacks.size(); i++) {
    if (in[i])
      continue;
    if (_game._packages[allpacks[i]]._weight +  weight > maxweight)
      continue;
    dist = abs(sum.x - _game._packages[allpacks[i]]._dest.x*numItems)
      + abs(sum.y - _game._packages[allpacks[i]]._dest.y*numItems);
    if (dist < mindist) {
      minindex = allpacks[i];
      mindist = dist;
    }
  }
  return minindex;

}

int
GoalStrategy::choose_random_package( const vector<unsigned> &allpacks,
    const vector<bool> &in, unsigned weight, unsigned maxweight)
{
  unsigned j;
  unsigned count = 0;
  j = rand_int(allpacks.size());
  while ( count < allpacks.size()  &&
      (in[j] == true 
      || (weight + _game._packages[allpacks[j]]._weight) > maxweight)) 
  {
    j = (j+1) % allpacks.size();
    count++;
  }
  if (count == allpacks.size())
    return -1;
  return allpacks[j];
}

void
GoalStrategy::complete_set(Point pos,
    const vector<unsigned> &allpacks,
    vector<bool> &in, unsigned &weight, unsigned maxweight,
    Point &sum, int numItems)
{
  vector<unsigned> weights;
  unsigned i,j,i_1;
  for (i = 0; i < allpacks.size(); i++) {
    if (in[i])
      continue;
    if (_game._packages[allpacks[i]]._dest == pos)
      weights.push_back(_game._packages[allpacks[i]]._weight);
  }

  sort(weights.begin(), weights.end());
  for (i = weights.size(); i > 0; i--) {
    i_1 = i - 1;
    if (weight + weights[i_1] <= maxweight) {
      for (j = 0; j < allpacks.size(); j++) {
	if ((in[j]==false) 
	    && _game._packages[allpacks[j]]._weight == weights[i_1]
	    && _game._packages[allpacks[j]]._dest == pos)
	{
	  in[j] = true;
	  weight += weights[i_1];
	  sum.x += pos.x;
	  sum.y += pos.y;
	  numItems++;
	  break;
	}
      }
    }
  }
}

