/*
 * $Id: Serial.h,v 1.1.1.1 2002/09/01 21:14:45 ben Exp $
 * Serial.h -- thom, 18 june 2001
 *     Data serialization.
 *
 */

#ifndef _SERIAL_H_
#define _SERIAL_H_

#include "ace/OS.h"
#include "Common.h"
#include <vector>
#include <list>


class DataOutputStream;
class DataInputStream;

#define MAX_VEC_SIZE 100000
#define MAX_STRING_SIZE 1024
#define MAX_BUF_SIZE 4096

/*
 * Serializable objects know how to read and write themselves on a stream.
 */
class SKY_EXPORT Serializable {
  public:
    // return value: 0 = success, negative = fail
    virtual int readObject(DataInputStream &s) = 0;
    virtual int writeObject(DataOutputStream &s) = 0;

    virtual ~Serializable() {}
};

/*
 * PolymorphicSerializable is a pattern for serializing and deserializing
 * a group of classes which all inherit from one main class.
 *
 * When it writes the object, it first writes the information about the
 * subtype that it is, then actually serializes the object.  Subclasses
 * of the superclass should inherit both from the superclass and from the
 * PolymorphicSerializable<superclass>.  
 *
 * The Superclass should have enumerated type OBJ_TYPE and it should 
 * implement the function createPolyObject with this signature:
 *
 *  static SUPERCLASS createSpecificObject(typename SUPERCLASS::OBJ_TYPE);
 *
 *  and these should be abstract, implemented in the subclasses:
 *  virtual typename SUPERCLASS::OBJ_TYPE get_object_type() = 0;
 *  virtual int readObject(DataInputStream &s) = 0;  // (Serializable)
 *  virtual int writeObject(DataOutputStream &s) = 0; // (Serializable)
 */
template<class SUPERCLASS>
class PolymorphicSerializable : public Serializable {
  public:

    static int writePolyObject 
	(SUPERCLASS* ptr, DataOutputStream &s) 
    {
	if (s.writeInt((int32)(ptr->get_object_type())) < 0) return -1;
	return ptr->writeObject(s);
    }
    static int readPolyObject 
	(SUPERCLASS* &ptr, DataInputStream &s)
    {
	int32 type;
	if (s.readInt(type) < 0) 
	    return -1;
	ptr = SUPERCLASS::createPolyObject((typename SUPERCLASS::OBJ_TYPE) type);
	if (ptr == NULL) 
	    return -1;
	if (ptr->readObject(s) < 0) {
	    delete ptr;
	    ptr = NULL;
	    return -1;
	   	
	}
	return 0;
    }
    virtual ~PolymorphicSerializable() {}
    PolymorphicSerializable() {}
};

/*
 * Data Streams
 * These give you the read and write classes for chars, ints etc
 * But they assume you have the smallest size.
 * i.e. char = 1, short = 2, int = 4, long = 4, 
 * float = 4, double = 8
 * If you're on a platform thats larger than this, it will cast down
 * If you're on a platform thats smaller than this then it will cast up 
 * but there may be data loss either way.
 *
 * The Var Length methods are when you don't know how large the data
 * will be, its just a protocol for writing the size of the array first.
 *
 * The dealloc booleans all over the place transfer which object is responsible
 * for deallocating a given buffer.
 */

/*
 * what about time-outs?
 * or other errors?
 * */
class DownloadPackage;

class DataOutputStream {
  public:
    // returns number of bytes written
    virtual int writeBuffer(void* buf, int len) = 0;

    // return status: 0 == successful, negative == error
    virtual int writeBool(bool b);
    virtual int writeChar(char ch);
    virtual int writeShort(short16 nShort);
    virtual int writeShort(ushort16 nShort);
    virtual int writeInt(int32 nInt);
    virtual int writeInt(uint32 nInt);
    virtual int writeLong(ulong64 nLong);
    virtual int writeString(const tstring &s);
    virtual int writeVectorString (vector<tstring> &v);
    virtual int writeVarLenBuffer(int32 nLen, char* buf, int timeouts = 1,
	    int * stop = NULL);
    virtual int writeWholeBuffer(void* buf, int len, int timeouts = 1,
	    int* stop = NULL);
    int writeObject(Serializable &ser);
    virtual void close() {}

#ifndef _WIN32
    template<class SERIALIZABLE>
    int writeVectorPtr(vector<SERIALIZABLE*> &v) {
	if (writeInt(v.size()) < 0)  return -1;
	vector<SERIALIZABLE*>::iterator i;
	for (i = v.begin(); i != v.end(); i++) {
	    if ((*i)->writeObject(*this) < 0) return -1;
	}
	return 0;
    }
#else
    int writeVectorPtr(vector<DownloadPackage*> &v);
#endif

    template<class SERIALIZABLE>
    int writeVector(vector<SERIALIZABLE> &v) {
	if (writeInt(v.size()) < 0)  return -1;
	vector<SERIALIZABLE>::iterator i;
	for (i = v.begin(); i != v.end(); i++) {
	    if ((*i).writeObject(*this) < 0) return -1;
	}
	return 0;
    }

    template<class SERIALIZABLE>
    int writeList(list<SERIALIZABLE> &l) {
	if (writeInt(l.size()) < 0)  return -1;
	list<SERIALIZABLE>::iterator i;
	for (i = l.begin(); i != l.end(); i++) {
	    if ((*i).writeObject(*this) < 0) return -1;
	}
	return 0;
    }

};

class DataInputStream {
  public:
    // returns number of bytes read
    virtual int readBuffer(void* buf, int len) = 0;

    // return status: 0 == successful, negative == error
    virtual int readBool(bool &b);
    virtual int readChar(char &buf);
    virtual int readChar(unsigned char &buf);
    virtual int readShort(short16 &buf);
    virtual int readShort(ushort16 &buf);
    virtual int readInt(int32 &buf);
    virtual int readInt(uint32 &buf);
    virtual int readLong(ulong64 &buf);
    virtual int readString(tstring &s, uint32 maxsize = MAX_STRING_SIZE); 
    virtual int readVectorString(vector<tstring> &v, 
	    uint32 maxvec = MAX_VEC_SIZE, uint32 maxstring = MAX_STRING_SIZE);
    virtual int readVarLenBuffer(uint32 &nLen, char* &buf,
	    uint32 maxsize = MAX_BUF_SIZE, int timeouts = 1, int* stop = NULL);
    virtual int readWholeBuffer(void* buf, int len, int timeouts = 1,
	    int* = NULL, int timeoutsInARowLimit = 5);
    virtual int readLine(char* line, int buflen);
    int readObject(Serializable &ser);

    virtual void close() {}

#ifndef _WIN32
    template<class SERIALIZABLE>
    int readVectorPtr(vector<SERIALIZABLE*> &v, 
	    uint32 maxsize = MAX_VEC_SIZE) {
	uint32 len;
	v.clear();
	bool success = true;
	if (readInt(len) < 0) 
	    return -1;
	if (len > maxsize) 
	    ERROR_RETURN((ERR, "read vec ptr -- vector too long %d", len), -1);
	for (unsigned int i = 0; i < len; i++) {
	    SERIALIZABLE *s =  new SERIALIZABLE();
	    if (s->readObject(*this)< 0) {
		delete s;
		success = false;
		break;
	    }
	    v.push_back(s);
	}
	if (success == false) {
	    vector<SERIALIZABLE*>::iterator j;
	    for (j = v.begin(); j != v.end(); j++) {
		delete (*j);
	    }
	    v.clear();
	    return -1;
	}
	return 0;
    }
#else
    int readVectorPtr(vector<DownloadPackage*> &v);
#endif


    template<class SERIALIZABLE>
    int readVector(vector<SERIALIZABLE> &v, uint32 maxsize = MAX_VEC_SIZE) {
	uint32 len;
	if (readInt(len) < 0) 
	    return -1;
	if (len > maxsize) {
	    ERROR_RETURN((ERR, "read vector - vector too long %d", len),-1);
	}
	v.resize(len);
	vector<SERIALIZABLE>::iterator i;
	for (i = v.begin(); i != v.end(); i++) {
	    if ((*i).readObject(*this) < 0) return -1;
	}
	return 0;
    }

    template<class SERIALIZABLE>
    int readList(list<SERIALIZABLE> &l, uint32 maxsize = MAX_VEC_SIZE) {
	uint32 len;
	if (readInt(len) < 0) 
	    return -1;
	if (len > maxsize) 
	    ERROR_RETURN((ERR, "readlist - list too long %d", len), -1);
	l.resize(len);
	list<SERIALIZABLE>::iterator i;
	for (i = l.begin(); i != l.end(); i++) {
	    if ((*i).readObject(*this) < 0) return -1;
	}
	return 0;
    }

};

class DataStream : public DataOutputStream, public DataInputStream
{
  public:
    // resolve ambiguity
    virtual void close() {
	DataInputStream::close();
	DataOutputStream::close();
    }
};

/*
 * File Stream
 */


class FileStream : public DataStream {
public:
    FileStream(FILE* fp);
    virtual int writeBuffer(void* buf, int len);
    virtual int readBuffer(void* buf, int len);

    virtual void close();
  private:
	FILE* m_fp;
};

/* 
 * Char Array Streams
 */ 

class CharArrayOutputStream : public DataOutputStream {
  private:
    char* m_buf;
    int m_nBufSize;
    uint32 m_nDataLen;

    bool m_ownArray;
  public:
    // writeBuffer returns the number of bytes written
    int writeBuffer(void* buffer, int len);
    void getCharArray (int &len, char* &buffer);
    void getCopyCharArray (int &len, char* &buffer);
    void getOwnershipCharArray(uint32 &len, char* &buffer);

    void reset();

    CharArrayOutputStream();
    CharArrayOutputStream(int start_size);
    virtual ~CharArrayOutputStream();
};

// CharArrayInputStream DOES own the buffer, and frees it.
// it's no longer going to take care of outside buffers at all.
#define CHARARRAYINPUTSTREAMSTARTSIZE	128
class CharArrayInputStream : public DataInputStream {
  private:
    char* m_buf;
    int m_nBufSize;
    int m_nDataPointer;
    uint32 m_nDataLen;
  public:
    void setCharArray (int nLen, void* buffer);
    int readBuffer(void* buf, int len);
    int getSize();

    CharArrayInputStream(int start_size = CHARARRAYINPUTSTREAMSTARTSIZE);
    CharArrayInputStream (int nLen, void* buffer);
    virtual ~CharArrayInputStream();
};

/* 
 * template methods
 * 
 */
/*
template<class SERIALIZABLE>
int
DataInputStream::readVector(vector<SERIALIZABLE> &v) {
    int len;
    if (readInt(len) < 0) return -1;
    v.resize(len);
    vector<SERIALIZABLE>::iterator i;
    for (i = v.begin(); i != v.end(); i++) {
        if ((*i).readObject(*this) < 0) return -1;
    }
    return 0;
}

template<class SERIALIZABLE>
int
DataInputStream::readList(list<SERIALIZABLE> &l) {
    int len;
    if (readInt(len) < 0) return -1;
    l.resize(len);
    list<SERIALIZABLE>::iterator i;
    for (i = l.begin(); i != l.end(); i++) {
        if ((*i).readObject(*this) < 0) return -1;
    }
    return 0;
}
template<class SERIALIZABLE>
int
DataOutputStream::writeVector(vector<SERIALIZABLE> &v) {
    if (writeInt(v.size()) < 0)  return -1;
    vector<SERIALIZABLE>::iterator i;
    for (i = v.begin(); i != v.end(); i++) {
        if ((*i).writeObject(*this) < 0) return -1;
    }
    return 0;
}
template<class SERIALIZABLE>
int
DataOutputStream::writeList(list<SERIALIZABLE> &l) {
    if (writeInt(l.size()) < 0)  return -1;
    list<SERIALIZABLE>::iterator i;
    for (i = l.begin(); i != l.end(); i++) {
        if ((*i).writeObject(*this) < 0) return -1;
    }
    return 0;
}

 
template <class SUPERCLASS>
int
PolymorphicSerializable<SUPERCLASS>::writePolyObject 
    (SUPERCLASS* ptr, DataOutputStream &s) 
{
    if (s.writeInt((int32)(ptr->get_object_type())) < 0) return -1;
    return ptr->writeObject(s);
}

template <class SUPERCLASS>
int
PolymorphicSerializable<SUPERCLASS>::readPolyObject 
    (SUPERCLASS* &ptr, DataInputStream &s)
{
    int32 type;
    if (s.readInt(type) < 0) 
        return -1;

    ptr = SUPERCLASS::createPolyObject((typename SUPERCLASS::OBJ_TYPE) type);
    if (ptr == NULL) 
        return -1;
    if (ptr->readObject(s) < 0) {
        delete ptr;
        ptr = NULL;
        return -1;
    }
    return 0;
}
*/
#endif
